/* (C) 2006 Richard Mills / The Linux Academy Ltd.
 *
 * Client side Tunnel.
 *
 * $Id: ClientTunnel.cc,v 1.28 2007-12-01 16:20:26 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "ClientTunnel.hh"
#include "Logger.hh"
#include "Debug.hh"
#include "MutexVar.hh"
#include "HTTPMsg.hh"
#include "TunnelConfig.hh"
#include "Authenticator.hh"
#include "GlobalConfig.hh"
#include "Connection.hh"
#include "HTTPConnection.hh"
#include "ClientTunneler.hh"
#include "PollTimeoutEvent.hh"
#include "SendMsgEvent.hh"


namespace HTTPTunnel {


// Global references
extern Logger* g_pLogger;
extern GlobalConfig* g_pConfig;


// Construction/destruction
ClientTunnel::ClientTunnel(TunnelConfig* tcfg, ClientTunneler* tunneler)
: Tunnel(tcfg, tunneler, RIGHT, CONNECT), m_pollTimeout(MIN_POLL_TIMEOUT),
  m_httpConnStable(false), m_nextPollTime(Time::now()), m_polling(false)
{
	DPRINTI("ClientTunnel::ClientTunnel()");

	// Start out non-persistant by default
	setPersistantMode(false);
}

ClientTunnel::~ClientTunnel()
{
	#ifdef DEBUG_TUN
	std::stringstream str;
	str << "ClientTunnel::~ClientTunnel() - " << m_tid;
	DPRINTI(str.str());
	#endif
}


/********************/
/* Public Interface */
/********************/


// Process some newly arrived tunnel msg from one of our connections
void ClientTunnel::processTunnelMsg(HTTPMsg* msg, HTTPConnection* httpConn)
{
	DPRINTV("ClientTunnel::processTunnelMsg() - " + getTidStr());

	MutexVar mv(m_mtx);

	// Check whether HTTP connection is clear
	const HTTPMsgParms* parms = msg->getParms();
	m_httpConnStable = (parms->getRespCode() == 200);

	// Update our quotas?
	if(parms->getState() == AUTHREQ) {
		Authenticator* authenticator = getAuthenticator();
		authenticator->setUserQuotas(m_user, parms->getQuotas());
	}

	// Do we have an indication of whether this connection is persistant?
	if(isNonPersistant()) {
		if(parms->getConnKeepAlive() || parms->getProxyConnKeepAlive()) {
			if(httpConn->getNoMsgsRecvd() > MAX_UNACKED_MSG) {
				#ifdef DEBUG_TUN
				std::stringstream dbg;
				dbg << "Tunnel " << getTidStr() << " recvd " << httpConn->getNoMsgsRecvd();
				dbg << " msgs over HTTP connection: Switching to persistant tunneling";
				DPRINTV(dbg.str());
				#endif

				// Looks like we can use persistant mode
				setPersistantMode(true);
			}
		}
	}

	// Base class
	Tunnel::processTunnelMsg(msg, httpConn);

	// See if poll timeout needs to change
	scalePollTimeout();

	// Do we have data waiting to send back to the server?
	if(m_rawDataQueueSize > 0) {
		// Check there isn't already a send msg pending
		unsigned int connID = httpConn->getID();
		if(!m_tunneler->isScheduled<SendMsgEvent, unsigned int>(connID)) {
			#ifdef DEBUG_TUN
			std::stringstream dbg;
			dbg << "ClientTunnel has " << m_rawDataQueueSize << " bytes queued, ";
			dbg << "scheduling early send";
			DPRINTV(dbg.str());
			#endif

			// Schedule msg send to server
			m_tunneler->scheduleEvent(new SendMsgEvent(Time(0,EARLY_MSG_SEND_TIME),
				connID, false, getTid(), m_tunneler));
		}
	}
}


// Poll the far end of the tunnel
void ClientTunnel::pollTunnel(void)
{
	DPRINTV("ClientTunnel::pollTunnel() - " + getTidStr());

	MutexVar mv(m_mtx);

	// Only send poll msg if we've been quiet
	Time quietTime = Time::now() - m_lastTunnelAnySendTime;
	if(quietTime.getTotalMillis() >= m_pollTimeout) {
		#ifdef DEBUG_TUN
		std::stringstream dbg;
		dbg << "Last msg sent " << quietTime << " ago, timeout = ";
		dbg << m_pollTimeout << " - polling";
		DPRINTV(dbg.str());
		#endif

		// Force the sending of a msg
		m_polling = true;
		sendTunnelMsg(true, 0);
		m_polling = false;
	}

	// Schedule next poll timeout
	m_nextPollTime = Time(0, m_pollTimeout);
	m_tunneler->scheduleEvent(new PollTimeoutEvent(m_nextPollTime, m_tunneler, m_tid));
}


// Set whether the tunnel user is authenticated
void ClientTunnel::setUserAuthenticated(bool auth)
{
	MutexVar mv(m_mtx);
	Authenticator* authenticator = getAuthenticator();
	authenticator->setUserAuthenticated(m_user, auth, this);
}


/***********************/
/* Protected Interface */
/***********************/


// Successfully sent a new message down the tunnel - Caller hold mutex
void ClientTunnel::sentNewTunnelMsg(HTTPMsgParms* parms)
{
	// We always want to remember in the client
	Tunnel::sentNewTunnelMsg(parms, true);
}


// Determine whether a mlti-ack implies we should resend a msg - Caller holds mutex
bool ClientTunnel::checkForMultiAckResendRequest(SentMsgParms* sentMsgParms)
{
	/* Since the server only responds to msgs we have sent, any duplicate
	 * acks imply a resend request.
	 */

	DPRINTV("ClientTunnel::checkForMultiAckResendRequest() - Multi-Ack resend request");
	return true;
}


// Have we got a stable HTTP connection to other end?
bool ClientTunnel::httpStable(void) const
{
	return m_httpConnStable;
}


// Can we send a new tunnel msg? - Caller holds mutex
bool ClientTunnel::clearToSendNewMsg(Connection* conn)
{
	bool sendNew = false;

	try {
		// Check for unacked management msgs and consider resending them early
		checkUnackedMgmtMsgs();

		// New message or resend of old message?
		sendNew = !resendUnackedMsgs(conn);
		if(sendNew) {
			// How many outstanding unacked msgs do we already have?
			if(m_unackedMsgQueue.size() >= m_maxUnackedMsg) {
				// If we're in a special connection state, go ahead with send anyway
				if(m_tunnelState == CONNECTED || m_tunnelState == CLOSING) {
					DPRINTI("Too many unacked msgs on tunnel, not sending another");
					sendNew = false;
				}
			}
		}
	}
	catch(ResendError& e) {
		// Resending error
		g_pLogger->warn("Error resending msg: " + e.what());
		recordConnectionError(MSG_RESEND_ERROR);
	}

	return sendNew;
}


// Check whether any connections need to be blocked/unblocked - Caller holds mutex
void ClientTunnel::checkBlockedConnections(void)
{
	// Check raw connections
	std::vector<Tunnel::Selector*> filt;
	filt.push_back(new SelLeftConn());
	std::vector<Connection*> conns;
	getConnsFromSel(filt, conns);

	// Do we want to block or unblock?
	bool unblock = false;
	bool block = (m_rawDataQueueSize >= MAX_RAW_DATA_FILL);
	if(!block) {
		// Only unblock if we're well below full
		unsigned int buffFill = (unsigned int)(((float)m_rawDataQueueSize /
			(float)MAX_RAW_DATA_FILL) * 100);
		unblock = (buffFill <= RAW_DATA_UNBLOCK_LEVEL);
	}

	// Check each connection
	std::vector<Connection*>::iterator cit;
	for(cit = conns.begin(); cit != conns.end(); ++cit) {
		// Get connection
		Connection* conn = *cit;

		// Need to change blocked state?
		if(block)
			conn->setBlocked(true);
		else if(unblock)
			conn->setBlocked(false);
	}
}


// Advance our internal state machine - Caller holds mutex
void ClientTunnel::advanceState(const HTTPMsgParms* parms)
{
	// Respond to the reported state from this msg
	switch(parms->getState()) {
		case(AUTHREQ): {
			// Become connected
			setState(CONNECTED);
			break;
		}
		case(CLOSING): {
			// Closing
			setState(CLOSING);
			break;
		}
		case(CLOSED): {
			// Closed
			setState(CLOSED);
			break;
		}
		case(CONNECTED): {
			// Stay connected
			setState(CONNECTED);
			break;
		}
		case(CONNECT):
		case(RECONNECT):
		case(UNKNOWNSTATE):
		default: {
			std::stringstream err;
			err << "Unexpected ClientTunnel reported msg state: " << parms->getState();
			g_pLogger->error(err.str());
			break;
		}
	}
}


// Get maximum allowable payload size - Caller holds mutex
unsigned int ClientTunnel::getMaxPayloadSize(bool useHeader)
{
	// Get payload size from client tunnel parameters
	ClientTunneler* cliTunneler = dynamic_cast<ClientTunneler*>(m_tunneler);
	ClientTunnelParms tunParms = cliTunneler->getTunParms();
	return tunParms.getMaxPayloadSize(useHeader);
}


// Something has been sent down the tunnel - Caller holds mutex
void ClientTunnel::tunnelAnyMsgSent(void)
{
	// Base class
	Tunnel::tunnelAnyMsgSent();

	// Reschedule any pending poll timeout
	if(!m_polling) {
		m_nextPollTime = Time(0, m_pollTimeout);
		m_tunneler->rescheduleEvent<PollTimeoutEvent, TunnelID>(m_tid, m_nextPollTime);
	}
}


// Scale poll timeout - Caller holds mutex
void ClientTunnel::scalePollTimeout(void)
{
	// Remember old poll timeout
	unsigned int oldPollTimeout = m_pollTimeout;

	// Set poll timeout at 10% below average ack rtt
	m_pollTimeout = m_biasAvgAckRTT - (m_biasAvgAckRTT / 10);

	// Check we're still within bounds
	if(m_pollTimeout > MAX_POLL_TIMEOUT)
		m_pollTimeout = MAX_POLL_TIMEOUT;
	else if(m_pollTimeout < MIN_POLL_TIMEOUT)
		m_pollTimeout = MIN_POLL_TIMEOUT;

	if(oldPollTimeout != m_pollTimeout) {
		#ifdef DEBUG_TUN
		std::stringstream nfo;
		nfo << "ClientTunnel::scalePollTimeout() - Now " << m_pollTimeout;
		nfo << " for tunnel " << getTidStr();
		DPRINTV(nfo.str());
		#endif

		// Work out new time for next poll timeout
		if(m_pollTimeout > oldPollTimeout) {
			// Reschedule for later
			m_nextPollTime.addInterval(0, (m_pollTimeout - oldPollTimeout));
		}
		else {
			// Reschedule for earlier
			m_nextPollTime.subInterval(0, (oldPollTimeout - m_pollTimeout));
		}

		// Reschedule poll timeout for new time
		m_tunneler->rescheduleEvent<PollTimeoutEvent, TunnelID>(m_tid, m_nextPollTime);
	}
}


}	// namespace

