/* (C) 2007 Richard Mills / The Linux Academy Ltd.
 *
 * HTTP Tunneler instance. Top level library interface.
 *
 * $Id: HTTPTunneler.cc,v 1.5 2008-02-23 14:01:40 rich Exp $
 *
 * This file is part of HTTPTunnel.
 *
 * HTTPTunnel is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * HTTPTunnel is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HTTPTunnel; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <iostream>
#include <sstream>
#include <fstream>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include "HTTPTunneler.hh"
#include "GlobalConfig.hh"
#include "OptionsParser.hh"
#include "Logger.hh"
#include "SSLSock.hh"
#include "Debug.hh"
#include "MutexVar.hh"


namespace HTTPTunnel {


/***************/
/* Global Data */
/***************/


// Global configuration settings
GlobalConfig* g_pConfig = 0;

// A logger
Logger* g_pLogger = 0;


#ifdef DEBUG_LOG_DATA
// Global log files + Mutex
std::ofstream* g_pRawReadLogFile = 0;
std::ofstream* g_pRawWriteLogFile = 0;
std::ofstream* g_pPayloadReadLogFile = 0;
std::ofstream* g_pPayloadWriteLogFile = 0;
unsigned int g_RawReadChars = 0;
unsigned int g_RawWriteChars = 0;
unsigned int g_PayloadReadChars = 0;
unsigned int g_PayloadWriteChars = 0;
Mutex g_logMtx;
#endif


/***************/
/* Static Data */
/***************/


// Singleton instance
HTTPTunneler* HTTPTunneler::c_tunneler = 0;
Mutex HTTPTunneler::c_mtx;


/********************/
/* Public Interface */
/********************/


/**********************************/
/* HTTPTunneler instance handling */
/**********************************/

// Get singleton instance
HTTPTunneler* HTTPTunneler::Get(void) throw(TunnelError)
{
	// Lock
	MutexVar mv(c_mtx);

	// Create singleton instance if required
	if(c_tunneler == 0) {
		c_tunneler = new HTTPTunneler();
	}

	// Return instance
	return c_tunneler;
}


// Release singleton instance
void HTTPTunneler::Release(void) throw(TunnelError)
{
	// Lock
	MutexVar mv(c_mtx);

	// Delete singleton instance if required
	if(c_tunneler) {
		delete c_tunneler;
		c_tunneler = 0;
	}
}


/*******************************/
/* Control the HTTP tunnelling */
/*******************************/

// Start the HTTP tunnelling in the background
void HTTPTunneler::Start(void) throw(TunnelError)
{
	// Take lock
	MutexVar mv(m_mtx);

	// Check we aren't already running
	if(!m_running) {
		#ifdef DEBUG_LOG_DATA
		// Open data debugging logs
		openDebugLogs();
		#endif

		// If requested, run as a daemon
		if(g_pConfig->getDaemon()) {
			mv.unlock();
			daemonise();
			mv.lock();
		}

		// Create any PID file
		createPIDFile();

		// Perform any final late configuration
		g_pConfig->doLateConfiguration();

		// Create configured tunnel servers from config
		g_pConfig->createLocalServers();

		// Indicate we are now running
		m_running = true;
	}
	else {
		throw TunnelError("HTTPTunneler already started");
	}
}


// Stop the HTTP tunnelling
void HTTPTunneler::Stop(void) throw(TunnelError)
{
	// Take lock
	MutexVar mv(m_mtx);

	// Check we are running
	if(m_running) {
		// Remove any PID file
		removePIDFile();

		// Destroy configured tunnel servers
		g_pConfig->destroyLocalServers();

		// Indicate we are no longer running
		m_running = false;
	}
}


/*********************************/
/* Configure the HTTP tunnelling */
/*********************************/

// Configure from a file
void HTTPTunneler::ConfigureFromFile(const char* fileName) throw(TunnelError)
{
	// Take lock
	MutexVar mv(m_mtx);

	// Parse configuration
	OptionsParser parser;
	parser.parseCfgFile(fileName);

	// Setup debug settings
	#ifdef DEBUG_TUN
	g_pConfig->setDaemon(false);
	#endif

	// For now, disable testing
	g_pConfig->setTesting(false);
}


// Regenerate the named config file with defaults
void HTTPTunneler::RegenerateCfgFile(const char* fileName) throw(TunnelError)
{
	// Take lock
	MutexVar mv(m_mtx);

	// Get parser to create default config options
	OptionsParser parser;
	parser.createDefaultConfigFile(fileName);
}


/*********************/
/* Private Interface */
/*********************/


// Construction
HTTPTunneler::HTTPTunneler() throw(TunnelError)
: m_mtx(), m_pidFileName(""), m_running(false)
{
	// Perform initialisation
	initialise();
}

// Destruction
HTTPTunneler::~HTTPTunneler() throw(TunnelError)
{
	// Take lock
	MutexVar mv(m_mtx);

	// Stop any active tunneling
	Stop();

	// Perform shutdown
	shutdown();
}


// Perform one-off initialisation
void HTTPTunneler::initialise(void) throw(TunnelError)
{
	// Construct logger
	#ifdef DEBUG_TUN
	g_pLogger = new Logger(STRNAME, true, false);
	#else
	g_pLogger = new Logger(STRNAME, false, true);
	#endif

	// Initialise SSL library
	try {
		SSLSock::initSSL();
	}
	catch(CommsError& e) {
		// Rethrow
		std::stringstream err;
		err << "HTTPTunneler initialisation failed: Error initialising SSL library: " << e.what();
		throw TunnelError(err.str());
	}

	// Construct default config
	g_pConfig = new GlobalConfig();
}


// Perform one-off shutdown
void HTTPTunneler::shutdown(void) throw(TunnelError)
{
	// Free globals
	if(g_pConfig) {
		delete g_pConfig;
		g_pConfig = 0;
	}
	if(g_pLogger) {
		delete g_pLogger;
		g_pLogger = 0;
	}

	// Shutdown SSL library
	try {
		SSLSock::shutdownSSL();
	}
	catch(CommsError& e) {
		// Rethrow
		std::stringstream err;
		err << "HTTPTunneler shutdown failed: Error closing SSL library: " << e.what();
		throw TunnelError(err.str());
	}
}


// Create a new PID file if requried - Caller holds mutex
void HTTPTunneler::createPIDFile(void) throw(TunnelError)
{
	// Tidy up any old PID file first
	removePIDFile();

	// New PID file requested?
	if(g_pConfig) {
		std::string pidfilename = g_pConfig->getPIDFile();
		if(pidfilename.length() > 0) {
			// Remember PID file name for tidying up
			m_pidFileName = pidfilename;

			// Create PID file
			std::ofstream pidfile(m_pidFileName.c_str());
			if(pidfile.good()) {
				pidfile << (unsigned int)getpid();
			}
			else {
				std::stringstream err;
				err << "HTTPTunneler failed to create PID file " << m_pidFileName;
				throw TunnelError(err.str());
			}
			pidfile.close();
		}
	}
}

// Tidy up a PID file if required - Caller holds mutex
void HTTPTunneler::removePIDFile(void)
{
	// Remove PID file
	if(m_pidFileName.length() > 0) {
		unlink(m_pidFileName.c_str());
		m_pidFileName = "";
	}
}


// Daemonise
void HTTPTunneler::daemonise(void) throw(TunnelError)
{
	pid_t pid = fork();
	if(pid < 0) {
		throw TunnelError("HTTPTunneler fork() error while daemonising");
	}
	else if(pid != 0) {
		// Parent just exits
		exit(0);
	}

	setsid();
	umask(0);
}


#ifdef DEBUG_LOG_DATA
// Open data debugging logs
void HTTPTunneler::openDebugLogs(void)
{
	// Set common filename stem
	std::string stem = g_pConfig->getDataLogFileStem();

	// Open log files
	std::string rawReadName = stem + "rawread.log";
	g_pRawReadLogFile = new std::ofstream(rawReadName.c_str());
	if(!g_pRawReadLogFile || !(*g_pRawReadLogFile)) {
		std::cerr << "Failed to open tunnel raw read log file" << std::endl;
		exit(1);
	}
	std::string rawWriteName = stem + "rawwrite.log";
	g_pRawWriteLogFile = new std::ofstream(rawWriteName.c_str());
	if(!g_pRawWriteLogFile || !(*g_pRawWriteLogFile)) {
		std::cerr << "Failed to open tunnel raw write log file" << std::endl;
		exit(1);
	}
	std::string payloadReadName = stem + "payloadread.log";
	g_pPayloadReadLogFile = new std::ofstream(payloadReadName.c_str());
	if(!g_pPayloadReadLogFile || !(*g_pPayloadReadLogFile)) {
		std::cerr << "Failed to open tunnel payload read log file" << std::endl;
		exit(1);
	}
	std::string payloadWriteName = stem + "payloadwrite.log";
	g_pPayloadWriteLogFile = new std::ofstream(payloadWriteName.c_str());
	if(!g_pPayloadWriteLogFile || !(*g_pPayloadWriteLogFile)) {
		std::cerr << "Failed to open tunnel payload write log file" << std::endl;
		exit(1);
	}
}
#endif


}	// namespace

